use serde::{Deserialize, Serialize};

// This is what Wolves returns to us
// success will always be 1?
#[derive(Deserialize, Serialize, Debug)]
struct WolvesResult<T> {
  success: i8,
  result: Vec<T>,
}

#[derive(Deserialize, Serialize, Debug)]
struct WolvesResultSingle<T> {
  success: i8,
  result: T,
}

/// Data returned for a member of a club/soc
#[derive(Deserialize, Serialize, Debug)]
pub struct WolvesUser {
  // TODO: Might be worth trying to get this replaced with the club/soc ID?
  /// Club/Soc the user is a member of
  pub committee: String,
  /// ID which uniquely identifies them on teh site
  pub member_id: String,
  /// First Name
  pub first_name: String,
  /// Last Name
  pub last_name: String,
  /// Email that they have set in their profile: <https://ulwolves.ie/memberships/profile>
  ///
  /// Note: This does not indicate if they wish to be contacted via this address.
  pub contact_email: String,
  /// Denotes if a user has opted into being emailed: ``0`` or ``1``
  pub opt_in_email: String,
  /// if the member is/was a student this is their ID
  pub student_id: Option<String>,
  /// Optional note set by Committee
  pub note: Option<String>,
  /// When their membership will expire: ``yyyy-MM-dd HH:mm:ss``
  pub expiry: String,
  /// When member requested membership: ``yyyy-MM-dd HH:mm:ss``
  pub requested: String,
  /// When the member was approved: ``yyyy-MM-dd HH:mm:ss``
  pub approved: String,
  /// Name of the site the user is a member of:``UL Wolves``
  pub sitename: String,
  /// Domain of the site they are a member of: ``ulwolves.ie``
  pub domain: String,
}

#[cfg(feature = "unstable")]
/// Information on an individual club/soc
#[derive(Deserialize, Serialize, Debug)]
pub struct WolvesCNS {
  /// ID of teh club/Society
  pub id: String,
  /// Name of the Club/Society
  pub name: String,
  /// Link to their page such as <https://ulwolves.ie/society/computer>
  pub link: String,
  /// Array of Committee members ``member_id````'s
  pub committee: Vec<String>,
}

#[derive(Serialize, Deserialize, Debug)]
#[serde(untagged)]
enum WolvesUserExists {
  B(bool),
  S(String),
}

pub struct Client {
  base_wolves: String,
  base_key: Option<String>,
}

impl Client {
  /// Create a new client for teh Wolves API
  ///
  /// # Arguments
  /// * `base_wolves` - Base URL for the requests, for example: ``https://cp.ulwolves.ie/api``
  /// * `base_key` - An instance API Key for running higher level privilege tasks
  ///
  /// # Examples
  ///
  /// ```rust
  /// use wolves_oxidised::Client;
  /// let client = Client::new("https://cp.ulwolves.ie/api", None);
  /// ```
  /// ```rust
  /// use wolves_oxidised::Client;
  /// let client = Client::new("https://cp.ulwolves.ie/api", Some("api_key"));
  /// ```
  pub fn new(base_wolves: &str, base_key: Option<&str>) -> Self {
    Self {
      base_wolves: base_wolves.to_string(),
      base_key: base_key.map(|x| x.to_string()),
    }
  }

  /// General method to get endpoints which return an array
  async fn get_bulk<T: serde::de::DeserializeOwned>(&self, wolves_endpoint: &str, api_key: &str) -> Vec<T> {
    if self.base_wolves.is_empty() {
      return vec![];
    }

    let url = format!("{}/{}", &self.base_wolves, wolves_endpoint);

    // get wolves data
    match reqwest::Client::new().post(&url).header("X-AM-Identity", api_key).send().await {
      Ok(x) => {
        if let Ok(WolvesResult {
          success,
          result,
        }) = x.json::<WolvesResult<T>>().await
        {
          if success != 1 {
            return vec![];
          }

          return result;
        }
      }
      Err(e) => {
        dbg!(e);
      }
    }
    vec![]
  }

  /// Get the members of a Club?Society based on teh API key inputted
  ///
  /// # Arguments
  /// * `api_key` - API key scoped to teh specific club/soc
  ///
  /// # Examples
  ///
  /// ```rust
  /// # tokio_test::block_on(async {
  /// use wolves_oxidised::{Client, WolvesUser};
  /// let client = Client::new("https://cp.ulwolves.ie/api", None);
  /// let result: Vec<WolvesUser> = client.get_members("api_key_club_1").await;
  /// # })
  /// ```
  pub async fn get_members(&self, api_key: &str) -> Vec<WolvesUser> {
    self.get_bulk::<WolvesUser>("get_members", api_key).await
  }

  #[cfg(feature = "unstable")]
  /// Get information about teh Club/Soc, including committee members
  ///
  /// # Examples
  /// No instance API key set
  /// ```rust
  /// # tokio_test::block_on(async {
  /// use wolves_oxidised::{Client, WolvesCNS};
  /// let client = Client::new("https://cp.ulwolves.ie/api", None);
  /// let result: Vec<WolvesCNS> = client.get_committees().await;
  /// assert_eq!(result.len(), 0);
  /// # })
  /// ```
  ///
  /// Instance API key set, will return details if there are no other errors
  /// ```rust
  /// # tokio_test::block_on(async {
  /// use wolves_oxidised::{Client, WolvesCNS};
  /// let client = Client::new("https://cp.ulwolves.ie/api", Some("api_key_instance"));
  /// let result: Vec<WolvesCNS> = client.get_committees().await;
  /// # })
  /// ```
  pub async fn get_committees(&self) -> Vec<WolvesCNS> {
    if let Some(api_key) = &self.base_key {
      self.get_bulk::<WolvesCNS>("get_cns", api_key).await
    } else {
      vec![]
    }
  }

  /// Get the ``member_id`` for a specific email if it exists.
  ///
  /// # Arguments
  /// * `api_key` - API key scoped to teh specific club/soc
  ///
  /// # Examples
  /// No instance API key set
  /// ```rust
  /// # tokio_test::block_on(async {
  /// use wolves_oxidised::Client;
  /// let client = Client::new("https://cp.ulwolves.ie/api", None);
  /// let result: Option<i64> = client.get_member("example@example.ie").await;
  /// assert!(result.is_none());
  /// # })
  /// ```
  ///
  /// Instance API key set, will return details if there are no other errors
  /// ```rust
  /// # tokio_test::block_on(async {
  /// use wolves_oxidised::Client;
  /// let client = Client::new("https://cp.ulwolves.ie/api", Some("api_key_instance"));
  /// let result: Option<i64> = client.get_member("example@example.ie").await;
  /// # })
  /// ```
  pub async fn get_member(self, email: &str) -> Option<i64> {
    // if the key isnt set then we cant do anything.
    let api_key = match &self.base_key {
      None => {
        return None;
      }
      Some(key) => key,
    };

    let url = format!("{}/get_id_from_email", &self.base_wolves);

    match reqwest::Client::new()
      .post(&url)
      .form(&[("email", email)])
      .header("X-AM-Identity", api_key)
      .send()
      .await
    {
      Ok(x) => {
        if let Ok(y) = x.json::<WolvesResultSingle<WolvesUserExists>>().await {
          // this is the only time we will get a positive response, the None at the end catches everything else
          if let WolvesUserExists::S(z) = y.result {
            if let Ok(id) = z.parse::<i64>() {
              return Some(id);
            }
          }
        }
      }
      Err(e) => {
        dbg!(e);
      }
    }

    None
  }
}

// pub fn add(left: u64, right: u64) -> u64 {
//     left + right
// }
//
// #[cfg(test)]
// mod tests {
//     use super::*;
//
//     #[test]
//     fn it_works() {
//         let result = add(2, 2);
//         assert_eq!(result, 4);
//     }
// }
